import subprocess
import pathlib
import os
import logging
import polars as pl
import shutil
import argparse
from multiprocessing import cpu_count

import os


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "300"
    os.environ["POLARS_FMT_STR_LEN"] = "100"


logging.basicConfig(
    level=logging.INFO,
    datefmt="%Y/%m/%d %H:%M:%S",
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def extract_filename(filepath: str) -> str:
    p = pathlib.Path(filepath)
    return p.stem


def generate_metric_file(
    bam_file: str,
    ref_fasta: str,
    out_filename: str,
    force: bool = False,
    threads=None,
    no_supp=False,
    no_mar=False,
) -> str:

    if no_supp and no_mar:
        raise ValueError("no_supp, no_mar can't be all true")

    if force and os.path.exists(out_filename):
        os.remove(out_filename)

    if os.path.exists(out_filename):
        logging.warning("use the existing metric file : %s", out_filename)
        return out_filename

    threads = cpu_count() if threads is None else threads
    cmd = f"""gsmm2-aligned-metric --threads {threads} \
            -q {bam_file} \
            -t {ref_fasta} \
            --out {out_filename} \
            --kmer 11 \
            --wins 1 """

    logging.info("cmd: %s", cmd)
    subprocess.check_call(cmd, shell=True)

    return out_filename


def analysis_segs(df: pl.DataFrame) -> pl.DataFrame:
    return (
        df.group_by(["segs"])
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [
                (pl.col("cnt") / pl.col("cnt").sum().over(pl.lit("1"))).alias("ratio"),
                pl.col("cnt").sum().over(pl.lit("1")).alias("tot_cnt"),
            ]
        )
        .sort("segs")
        .select(
            [
                pl.format("[SegsRatio]segs={}", pl.col("segs")).alias("name"),
                pl.col("ratio").alias("value"),
            ]
        )
    )


def analysis_segs2(df: pl.DataFrame) -> pl.DataFrame:
    return (
        df.filter(pl.col("segs") == pl.lit(2))
        .with_columns(
            [
                (pl.col("qOvlpRatio") < 0.01).alias("nonOvlpQuery"),
            ]
        )
        .with_columns(
            [
                (pl.col("nonOvlpQuery").and_(pl.col("rOvlpRatio") > 0.90)).alias(
                    "svCandidate"
                ),
                (pl.col("nonOvlpQuery").and_(pl.col("rOvlpRatio") < 0.01)).alias(
                    "noCutCandidate"
                ),
            ]
        )
        .with_columns(
            [pl.col("oriQGaps").str.split(",").list.get(1).cast(pl.Int32).alias("gap")]
        )
        .with_columns([pl.col("gap").lt(pl.lit(20)).alias("gap<20")])
        .with_columns(
            [
                pl.col("svCandidate").and_(pl.col("gap<20")).alias("sv"),
                pl.col("noCutCandidate").and_(pl.col("gap<20")).alias("noCut"),
            ]
        )
        .with_columns(
            [
                pl.when(pl.col("sv"))
                .then(pl.lit("sv"))
                .when(pl.col("noCut"))
                .then(pl.lit("noCut"))
                .when(pl.col("svCandidate"))
                .then(pl.lit("svCandidate"))
                .when(pl.col("noCutCandidate"))
                .then(pl.lit("noCutCandidate"))
                .otherwise(pl.lit("badCase"))
                .alias("tag")
            ]
        )
        .group_by(["tag"])
        .agg([pl.len().alias("cnt")])
        .select(
            [
                pl.col("tag"),
                pl.col("cnt"),
                pl.col("cnt").sum().over(pl.lit("1")).alias("tot_cnt"),
            ]
        )
        .with_columns([(pl.col("cnt") / pl.col("tot_cnt")).alias("ratio")])
        .sort(["tag"])
        .select(
            [
                pl.format("[segs=2]{}", pl.col("tag")).alias("name"),
                pl.col("ratio").alias("value"),
            ]
        )
    )


def stats(metric_filename, filename):
    df = pl.read_csv(metric_filename, separator="\t").filter(pl.col("rname") != "")
    # print(df.head(2))
    # print(
    #     df.filter(pl.col("segs") > 1)
    #     .head(2)
    #     .select(
    #         [
    #             "qname",
    #             "rname",
    #             "qlen",
    #             "segs",
    #             "queryCoverage",
    #             "identity",
    #             "oriAlignInfo",
    #             "oriQGaps",
    #             "qOvlp",
    #             "qOvlpRatio",
    #             "rOvlpRatio",
    #             "mergedQrySpan",
    #         ]
    #     )
    # )
    metric_segs = analysis_segs(df)
    metric_segs2 = analysis_segs2(df)

    df = df.with_columns(
        [
            ((pl.col("qOvlpRatio") < 0.01).and_(pl.col("rOvlpRatio") < 0.01))
            .or_(
                (pl.col("qOvlpRatio") < 0.01)
                .and_(pl.col("rOvlpRatio") > 0.90)
                .and_(pl.col("oriQGaps").str.split(",").list.get(1).cast(pl.Int32) < 20)
            )
            .alias("valid")
        ]
    )

    df = df.with_columns(
        [
            pl.when(pl.col("valid"))
            .then(pl.col("covlen"))
            .otherwise(pl.col("primaryCovlen"))
            .alias("miscCovlen")
        ]
    )

    df = df.with_columns(row_align_span())
    aggr_metrics = df.select(aggr_expressions())
    aggr_metrics = aggr_metrics.transpose(
        include_header=True, header_name="name", column_names=["value"]
    )

    aggr_metrics = pl.concat([aggr_metrics, metric_segs, metric_segs2])

    print(aggr_metrics)

    aggr_metrics.write_csv(filename, include_header=True, separator="\t")


def aggr_expressions():

    exprs = [
        (pl.col("primaryCovlen").sum() / pl.col("qlen").sum()).alias("queryCoverage"),
        (pl.col("miscCovlen").sum() / pl.col("qlen").sum()).alias("queryCoverage2"),
        (pl.col("covlen").sum() / pl.col("qlen").sum()).alias("queryCoverage3"),
        (pl.col("match").sum() / pl.col("alignSpan").sum()).alias("identity"),
        (pl.col("misMatch").sum() / pl.col("alignSpan").sum()).alias("mmRate"),
        (pl.col("ins").sum() / pl.col("alignSpan").sum()).alias("NHInsRate"),
        (pl.col("homoIns").sum() / pl.col("alignSpan").sum()).alias("HomoInsRate"),
        (pl.col("del").sum() / pl.col("alignSpan").sum()).alias("NHDelRate"),
        (pl.col("homoDel").sum() / pl.col("alignSpan").sum()).alias("HomoDelRate"),
    ]

    for base in "ACGT":
        exprs.extend(
            [
                (
                    pl.col(f"match-{base}").sum() / pl.col(f"alignSpan-{base}").sum()
                ).alias(f"identity-{base}"),
                (
                    pl.col(f"misMatch-{base}").sum() / pl.col(f"alignSpan-{base}").sum()
                ).alias(f"mmRate-{base}"),
                (pl.col(f"ins-{base}").sum() / pl.col(f"alignSpan-{base}").sum()).alias(
                    f"NHInsRate-{base}"
                ),
                (
                    pl.col(f"homoIns-{base}").sum() / pl.col(f"alignSpan-{base}").sum()
                ).alias(f"HomoInsRate-{base}"),
                (pl.col(f"del-{base}").sum() / pl.col(f"alignSpan-{base}").sum()).alias(
                    f"NHDelRate-{base}"
                ),
                (
                    pl.col(f"homoDel-{base}").sum() / pl.col(f"alignSpan-{base}").sum()
                ).alias(f"HomoDelRate-{base}"),
            ]
        )

    return exprs


def row_align_span():
    exprs = [
        (
            pl.col("match")
            + pl.col("misMatch")
            + pl.col("ins")
            + pl.col("homoIns")
            + pl.col("del")
            + pl.col("homoDel")
        ).alias("alignSpan")
    ]

    for base in "ACGT":
        exprs.append(
            (
                pl.col(f"match-{base}")
                + pl.col(f"misMatch-{base}")
                + pl.col(f"ins-{base}")
                + pl.col(f"homoIns-{base}")
                + pl.col(f"del-{base}")
                + pl.col(f"homoDel-{base}")
            ).alias(f"alignSpan-{base}")
        )
    return exprs


def main(bam_file: str, ref_fa: str, threads=None, force=False, outdir=None) -> str:
    """
        step1: generate detailed metric info
        step2: compute the aggr metric. the result aggr_metric.csv is a '\t' seperated csv file. the header is name\tvalue
            here is a demo.
            ---------aggr_metric.csv
            name    value
            queryCoverage   0.937
            ----------

    requirements:
        mm2: cargo install mm2 (version >= 0.19.0)

    Args:
        bam_file (str): bam file. only support adapter.bam
        ref_fa (str): ref genome fa file nam
        threads (int|None): threads for generating detailed metric file
        force (boolean): if force==False, use the existing metric file if exists
        outdir: if None, ${bam_filedir}/${bam_file_stem}-metric as outdir

    Return:
        (aggr_metric_filename, fact_metric_filename) (str, str)
    """
    bam_filedir = os.path.dirname(bam_file)
    stem = extract_filename(bam_file)
    if outdir is None:
        outdir = os.path.join(bam_filedir, f"{stem}-metric")

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    fact_metric_filename = f"{outdir}/{stem}.gsmm2_aligned_metric_fact.csv"
    fact_metric_filename = generate_metric_file(
        bam_file,
        ref_fa,
        out_filename=fact_metric_filename,
        force=force,
        threads=threads,
    )
    aggr_metric_filename = f"{outdir}/{stem}.gsmm2_aligned_metric_aggr.csv"
    if force and os.path.exists(aggr_metric_filename):
        os.remove(aggr_metric_filename)

    # if not os.path.exists(aggr_metric_filename):
    stats(fact_metric_filename, filename=aggr_metric_filename)
    # else:
    #     logging.warning(
    #         "aggr_metric_file exists, use existing one. %s", aggr_metric_filename
    #     )
    return (aggr_metric_filename, fact_metric_filename)


def test_stat():
    fact_bam_basic = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/fact_aligned_bam_bam_basic.csv"
    aggr_metric_filename = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/aggr_metric.csv"

    with open(aggr_metric_filename, encoding="utf8", mode="w") as file_h:
        file_h.write(f"name\tvalue\n")
        stats(fact_bam_basic, file_h=file_h)


def sv_identification(ori_align_info: str):
    """structural variant identification
    rule:
        small ovlp between aligned segments
        different segments align to the different reference regions
    """
    if ";" not in ori_align_info:
        return False

    align_regions = ori_align_info.split(";")
    align_regions = [align_region[:-1] for align_region in align_regions]
    # align_regions = [align_region.split(":")]
    pass


def adapter_remover_error_identification():
    """adapter remover error identification

    rule:
        small ovlp between aligned segments
        different segments align to the similar reference region

        if gap < 10: treat as adapter_missing
        if gap > 10: treat as adapter_lowq

    """
    pass


if __name__ == "__main__":
    polars_env_init()

    parser = argparse.ArgumentParser(prog="parser")
    parser.add_argument("--bams", nargs="+", type=str, required=True)
    parser.add_argument("--refs", nargs="+", type=str, required=True)
    parser.add_argument("-f", action="store_true", default=False)

    args = parser.parse_args()

    bam_files = args.bams
    refs = args.refs
    if len(refs) == 1:
        refs = refs * len(bam_files)

    assert len(bam_files) == len(refs)

    # bam_files = [
    #     "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter.bam",
    #     # "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0002_adapter.bam",
    #     # "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0003_adapter.bam",
    # ]
    # ref = "/data/ccs_data/MG1655.fa"

    for bam, ref in zip(bam_files, refs):
        main(bam_file=bam, ref_fa=ref, force=args.f)
    # test_stat()

    # print(merge_intervals([{"qstart": 11, "qend": 771}]))



import polars as pl
import os
import sys

cur_dir = os.path.abspath(__file__).rsplit("/", maxsplit=1)[0]
sys.path.insert(0, cur_dir)

import polars_init
import argparse


def variant_calling(df: pl.DataFrame):
    v = (
        df.with_columns([(pl.col("eq") + pl.col("diff")).alias("eq_and_diff")])
        .with_columns(
            [
                (pl.col("eq") / pl.col("eq_and_diff")).alias("eq_in_eqdiff"),
                (pl.col("diff") / pl.col("eq_and_diff")).alias("diff_in_eqdiff"),
            ]
        )
        .filter((pl.col("eq_in_eqdiff") > 0.2).and_(pl.col("diff_in_eqdiff") > 0.2))
        .filter(pl.col("diffDetail").str.split(",").list.len() < 2)
        .shape[0]
    )

    print("variant calling ratio: {} / {} = {}".format(v, df.shape[0], v / df.shape[0]))


def del_calling(df: pl.DataFrame):
    v = (
        df.with_columns([(pl.col("del") / pl.col("depth")).alias("del_ratio")])
        .filter((pl.col("del_ratio") > 0.4))
        .shape[0]
    )

    print("del ratio: {} / {} = {}".format(v, df.shape[0], v / df.shape[0]))


def ana(filepath: str):
    df = pl.read_csv(filepath, separator="\t")
    print(df.head(2))

    df = (
        df.with_columns(
            [
                (
                    pl.col("eq")
                    / (pl.col("eq") + pl.col("diff") + pl.col("ins") + pl.col("del"))
                ).alias("eq_rate"),
                (pl.col("eq") / pl.col("depth")).alias("eq_rate2"),
            ]
        )
        # .filter(pl.col("curIsHomo").eq(0).and_(pl.col("nextIsHomo").eq(0)))
        .sort(by=["eq_rate2"], descending=[False])
    )

    print(df.head(100))

    print(df.select((pl.col("eq_rate2") < 0.5).sum() / pl.len()))

    variant_calling(df=df)
    del_calling(df=df)

    # print(df.head(200))


if __name__ == "__main__":
    polars_init.polars_env_init()

    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("fp", metavar="fact_aligned_bam_ref_locus_info.csv")
    args = parser.parse_args()
    ana(filepath=args.fp)
    pass



import pysam
import argparse
from tqdm import tqdm


def calculate_long_insertions(bam_file, min_insertion_length):
    # 打开 BAM 文件
    bam = pysam.AlignmentFile(bam_file, "rb", threads=40)

    total_queries = 0
    queries_with_long_insertions = 0
    total_query_length = 0
    total_long_insertion_length = 0

    for read in tqdm(bam.fetch(), desc=f"reading {bam_file}"):
        # 统计总 query 数量
        total_queries += 1

        if read.is_unmapped:
            continue  # 跳过未比对的 reads

        query_length = read.query_length or 0
        total_query_length += query_length

        # 遍历 CIGAR 操作
        has_long_insertion = False
        long_insertion_length = 0

        for op, length in read.cigartuples or []:
            if op == 1 and length >= min_insertion_length:  # 插入 (CIGAR code 1)
                has_long_insertion = True
                long_insertion_length += length

        if has_long_insertion:
            queries_with_long_insertions += 1
            total_long_insertion_length += long_insertion_length

    bam.close()

    # 计算统计结果
    proportion_queries_with_long_insertions = (
        queries_with_long_insertions / total_queries if total_queries > 0 else 0
    )
    proportion_long_insertion_length = (
        total_long_insertion_length / total_query_length
        if total_query_length > 0
        else 0
    )

    return {
        "total_queries": total_queries,
        "queries_with_long_insertions": queries_with_long_insertions,
        "proportion_queries_with_long_insertions": proportion_queries_with_long_insertions,
        "total_query_length": total_query_length,
        "total_long_insertion_length": total_long_insertion_length,
        "proportion_long_insertion_length": proportion_long_insertion_length,
    }


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="BAM file long insertion statistics.")
    parser.add_argument(
        "-b", "--bam", required=True, help="Input BAM file (sorted and indexed)."
    )
    parser.add_argument(
        "-l",
        "--length",
        type=int,
        default=10,
        help="Minimum length to consider an insertion as long (default: 10).",
    )
    args = parser.parse_args()

    stats = calculate_long_insertions(args.bam, args.length)

    print("BAM File: ", args.bam)
    print("Total Queries: ", stats["total_queries"])
    print("Queries with Long Insertions: ", stats["queries_with_long_insertions"])
    print(
        "Proportion of Queries with Long Insertions: {:.2%}".format(
            stats["proportion_queries_with_long_insertions"]
        )
    )
    print("Total Query Length: ", stats["total_query_length"])
    print("Total Long Insertion Length: ", stats["total_long_insertion_length"])
    print(
        "Proportion of Long Insertion Length: {:.2%}".format(
            stats["proportion_long_insertion_length"]
        )
    )



import pathlib
import os
import subprocess
import sys
import pysam
from tqdm import tqdm
import polars as pl
import argparse

pre_dir = os.path.abspath(__file__).rsplit("/", maxsplit=2)[0]
print(pre_dir)
sys.path.append(pre_dir)

from fact_table_ana.utils import read_fastx_file
from fact_table_ana.polars_init import polars_env_init


def do_alignment(hpc_bam: str, hpc_ref: str):
    out_root = os.path.dirname(hpc_bam)

    o_path = "{}/{}-{}.aligned".format(
        out_root, pathlib.Path(hpc_bam).stem, pathlib.Path(hpc_ref).stem
    )

    cmd = f"gsmm2 align -q {hpc_bam} -t {hpc_ref} --noMar -p {o_path}"
    subprocess.check_call(cmd, shell=True)
    return f"{o_path}.bam"


def read_bam_data(aligned_bam: str, hpc_ref_cnt: str, hpc_ref_seq: str):
    ref_hpc_seq_data = read_fastx_file(hpc_ref_seq)

    ref_hpc_cnt_data = read_fastx_file(hpc_ref_cnt)
    ref_hpc_cnt_data = {
        qname: list(map(int, seq[0].split(",")))
        for qname, seq in ref_hpc_cnt_data.items()
    }

    counter = {}

    with pysam.AlignmentFile(aligned_bam, mode="rb", threads=40) as bam_h:
        for refname in ref_hpc_cnt_data.keys():
            ref_seq = ref_hpc_seq_data[refname][0]
            ref_hpc_cnt = ref_hpc_cnt_data[refname]

            for i, record in enumerate(
                tqdm(bam_h.fetch(refname), desc=f"reading {refname}")
            ):
                if i > 50000:
                    break
                aligned_pairs = record.get_aligned_pairs(matches_only=True)
                query_hpc_cnt = record.get_tag("dw")
                query_seq = record.query_sequence
                for qpos, rpos in aligned_pairs:
                    if query_seq[qpos] != ref_seq[rpos]:
                        continue

                    if ref_hpc_cnt[rpos] <= 1:
                        continue
                    counter.setdefault(
                        f"{ref_seq[rpos]}{ref_hpc_cnt[rpos]}", []
                    ).append(query_hpc_cnt[qpos])
    keys = []
    values = []
    for k, counts in counter.items():
        for c in counts:
            keys.append(k)
            values.append(c)

    return pl.DataFrame({"key": keys, "value": values})


def ana(df: pl.DataFrame):
    df = (
        df.group_by(["key", "value"])
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [
                (pl.col("cnt") / pl.col("cnt").sum().over(pl.col("key"))).alias(
                    "ratio"
                ),
                pl.col("key").str.slice(1).cast(pl.Int32).alias("ref_hp_cnt"),
            ]
        )
        .filter((pl.col("ref_hp_cnt") - pl.col("value")).abs() <= 1)
    ).sort(by=["key", "value"])

    df = df.with_columns(
        [
            (pl.col("ref_hp_cnt") + 1).alias("ref_hp_cnt_ADD_1"),
            (pl.col("ref_hp_cnt") - 1).alias("ref_hp_cnt_MINUS_1"),
        ]
    )

    df_eq = df.filter(pl.col("ref_hp_cnt") == pl.col("value"))
    df_minus_one = df.filter(pl.col("ref_hp_cnt_MINUS_1") == pl.col("value"))
    df_add_one = df.filter(pl.col("ref_hp_cnt_ADD_1") == pl.col("value"))

    all = df_eq.join(df_minus_one, on="key", suffix="_minus_one").join(
        df_add_one, on="key", suffix="_add_one"
    )

    print(all.head(1000))


def main(args):

    # align_res = do_alignment(args.hpc_bam, args.hpc_ref)
    align_res = "/data/ccs_data/ccs_eval2024q3/jinpu/20240711_Sync_Y0006_02_H01_Run0001_called.subreads.hpc-ref_Saureus_ATCC25923.m.new.corrected.hpc.aligned.bam"
    df = read_bam_data(align_res, args.hpc_ref_cnt, args.hpc_ref)
    ana(df=df)

    pass


if __name__ == "__main__":
    polars_env_init()

    cli_params = {
        "hpc_bam": "/data/ccs_data/ccs_eval2024q3/jinpu/20240711_Sync_Y0006_02_H01_Run0001_called.subreads.hpc.bam",
        "hpc_ref": "/data/ccs_data/ccs_eval2024q3/jinpu/ref_Saureus_ATCC25923.m.new.corrected.hpc.fa",
        "hpc_ref_cnt": "/data/ccs_data/ccs_eval2024q3/jinpu/ref_Saureus_ATCC25923.m.new.corrected.hpc.txt",
    }

    main(argparse.Namespace(**cli_params))



import os
import sys

cur_dir = os.path.abspath(__file__).rsplit("/", maxsplit=1)[0]
sys.path.insert(0, cur_dir)
sys.path.insert(0, ".")
import polars as pl
import argparse
import polars_init
import numpy as np


def analysis(fact_poly_info_file: str):

    df = pl.read_csv(fact_poly_info_file, separator="\t")
    df = df.filter(pl.col("qClean")).with_columns(
        [pl.format("{}{}", pl.col("rBase"), pl.col("rRepeats")).alias("key")]
    )

    df = (
        df.group_by(["key", "qRepeats"])
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [
                (pl.col("cnt") / pl.col("cnt").sum().over(pl.col("key"))).alias(
                    "ratio"
                ),
                pl.col("key").str.slice(1).cast(pl.Int32).alias("ref_hp_cnt"),
            ]
        )
    ).sort(by=["key", "qRepeats"])

    df = df.with_columns(
        [
            (pl.col("ref_hp_cnt") + 1).alias("ref_hp_cnt_ADD_1"),
            (pl.col("ref_hp_cnt") - 1).alias("ref_hp_cnt_MINUS_1"),
        ]
    )

    df_eq = df.filter(pl.col("ref_hp_cnt") == pl.col("qRepeats"))
    df_minus_one = df.filter(pl.col("ref_hp_cnt_MINUS_1") == pl.col("qRepeats"))
    df_add_one = df.filter(pl.col("ref_hp_cnt_ADD_1") == pl.col("qRepeats"))

    all = (
        df_eq.join(df_minus_one, on="key", suffix="_minus_one", how="left")
        .join(df_add_one, on="key", suffix="_add_one", how="left")
        .select(
            [
                "key",
                "qRepeats",
                "cnt",
                "ratio",
                "qRepeats_minus_one",
                "cnt_minus_one",
                "ratio_minus_one",
                "qRepeats_add_one",
                "cnt_add_one",
                "ratio_add_one",
            ]
        )
    )
    print(all.head(100))

    print(df.select(["key", "qRepeats", "cnt", "ratio"]).sort(by=["key", "qRepeats"]))
    return df


def map(model: pl.DataFrame, base, cnts):
    results = np.array([-100] * 10)
    for n in range(0, len(results)):

        for cnt in cnts:
            prob = model.filter(
                (pl.col("key") == f"{base}{n}").and_(pl.col("qRepeats") == cnt)
            ).select(pl.col("ratio").log())
            if prob.shape[0] == 0:
                results[n] += -100
            else:
                prob = float(prob.to_pandas()["ratio"])
                results[n] += prob
    results = np.exp(results)
    print(results)
    print(results.argmax())
    pass


def main(args):
    df = analysis(args.fact_poly_info_file)
    return

    examples = [
        ["A", [3, 4, 4, 3, 3, 4, 4, 3, 4]],
        ["A", [4, 4, 3, 3, 3, 3, 3]],
        ["T", [6, 5, 5, 3, 6, 3, 4, 4]],
        ["A", [5, 3, 5, 5, 4, 4, 3, 5]],
        ["T", [2, 2, 2, 2, 2, 4, 5]],
        ["A", [6, 6, 6, 5, 6, 5, 6]],
        ["A", [3, 2, 3, 5, 3, 4, 4]],
        ["T", [5, 3, 5, 4, 4, 6, 5]],
        ["A", [3, 3, 4, 4, 3, 4, 4, 4]],
        ["T", [4, 4, 3, 4, 4, 2, 4]],
        ["A", [2, 2, 3, 2, 2, 3]],
        ["T", [6, 5, 6, 6, 5, 5, 6, 6]],
        ["A", [3, 2, 3, 2, 3, 3, 2]],
        ["G", [2, 2, 1, 3, 3, 2, 3, 2, 2, 2]],
        ["T", [6, 5, 6, 5, 6, 6, 6]],
    ]
    for example in examples:
        map(df, example[0], example[1])

    pass


if __name__ == "__main__":
    polars_init.polars_env_init()
    cli_params = {
        "fact_poly_info_file": "/data/adam/20241203_240901Y0005_Run0001/analysis/fact_poly_info.csv"
    }
    main(argparse.Namespace(**cli_params))



import pysam
import os
import sys
from tqdm import tqdm
import argparse

pre_dir = os.path.abspath(__file__).rsplit("/", maxsplit=2)[0]
sys.path.append(pre_dir)


def insert_cs_to_sbr_bam(sbr_bam: str, smc_bam: str):
    out_bam = "{}.with-cs.bam".format(sbr_bam.rsplit(".", maxsplit=1)[0])

    with pysam.AlignmentFile(sbr_bam, mode="rb", threads=40, check_sq=False) as in_h:
        with pysam.AlignmentFile(
            out_bam, mode="wb", threads=40, check_sq=False, header=in_h.header
        ) as out_h:
            for record in tqdm(in_h.fetch(until_eof=True), desc=f"dumping1 {out_bam}"):
                out_h.write(record)

            with pysam.AlignmentFile(
                smc_bam, mode="rb", threads=40, check_sq=False
            ) as in_h:
                for in_record in tqdm(
                    in_h.fetch(until_eof=True), desc=f"dumping2 {out_bam}"
                ):
                    qname = in_record.query_name
                    record = pysam.AlignedSegment()
                    record.query_name = f"00_{qname}"
                    record.query_sequence = in_record.query_sequence
                    record.set_tag("ch", int(in_record.get_tag("ch")), value_type="I")
                    out_h.write(record)


def main(args):
    insert_cs_to_sbr_bam(args.sbr_bam, args.smc_bam)
    pass


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("--sbr-bam", dest="sbr_bam")
    parser.add_argument("--smc-bam", dest="smc_bam")

    main(args=parser.parse_args())



"""
this script is used to visualize the msa.
the result can be used in two ways:
    * copy the fasta info and past to the jalview, jalview is used for the plot. OR
    * generate the msa picture directly

the whole pipeline is:
1) do alignment (pairwise alignment)  mini_align -P -m -r ref_filepath.fa -i inp.fa -t 1 -p calls2draft
2) use this script to generate msa pic

"""

import argparse
import pysam
import numpy as np
import tempfile
import os
from typing import Mapping
from tqdm import tqdm


def read_fastx_file(fname: str):
    fh = pysam.FastxFile(fname)
    res = {}
    for entry in fh:
        res[entry.name] = entry.sequence
    return res


def read_bam_file(bam_file: str) -> Mapping[str, str]:
    res = {}
    with pysam.AlignmentFile(bam_file, mode="rb", threads=40, check_sq=False) as bam_h:
        for record in tqdm(
            bam_h.fetch(until_eof=True), desc=f"read_bam_file:>> reading {bam_file}"
        ):
            res[record.query_name] = record.query_sequence

    return res


class Name2Seq:
    def __init__(self, fname: str):
        if (
            fname.endswith("fa")
            or fname.endswith("fasta")
            or fname.endswith("fna")
            or fname.endswith("fq")
            or fname.endswith("fastq")
        ):
            self.qname2seq = read_fastx_file(fname)
        elif fname.endswith("bam"):
            self.qname2seq = read_bam_file(fname)
        else:
            raise ValueError(f"invalid file format, {fname}")

    def fetch(self, name):
        return self.qname2seq[name]


def build_query_name(align_seg: pysam.AlignedSegment):
    return f"{align_seg.query_name}_SE_{align_seg.query_alignment_start}_{align_seg.query_alignment_end}"


class ResultMatrix:
    """msa alignment matrix"""

    def __init__(self, ref_start, ref_end, query_names, ref_name=None) -> None:
        ref_len = ref_end - ref_start
        self.num_records = len(query_names) + 1  # +ref
        self.query_names = sorted(query_names)
        self.query2idx = {
            query_name: idx for idx, query_name in enumerate(self.query_names, start=1)
        }
        self.ref_name = "REF" if ref_name is None else ref_name
        self.ref_name = f"{self.ref_name}_{ref_start}_{ref_end}"
        self.matrix = ResultMatrix.init_matrix(ref_len * 2, self.num_records)

        self.ref_end = ref_end  # exclusive
        self.ref_start = ref_start

        self.cur_row = 0
        self.cur_ref_pos = self.ref_start
        self.first_query_is_rev = None

    def update(self, pileup_col: pysam.PileupColumn, ref: str = None):
        """
        query1 query2 query3 , ....

        A        A
        .        A

        """
        self.cur_ref_pos = pileup_col.reference_pos

        max_ins = 0
        position_names = []
        for query in pileup_col.pileups:
            position_names.append(query.alignment.query_name)
            if query.indel > 0:
                max_ins = max([max_ins, query.indel])
        # print(self.cur_ref_pos, "\n", "\n".join(sorted(position_names)))

        self.extend_matrix(max_offset=max_ins + 1)
        if ref is not None:
            self.matrix[self.cur_row, 0] = ref[pileup_col.reference_pos]

        # print(self.cur_ref_pos, len(pileup_col.pileups))
        for query in pileup_col.pileups:
            if query.is_refskip:
                continue
            cur_query_max_offset = max([0, query.indel])
            q_name = build_query_name(query.alignment)
            query_pos = self.query2idx[q_name]
            if query_pos == 1:
                self.first_query_is_rev = query.alignment.is_reverse
                # print(f"{query.alignment.query_name} --> {query.alignment.is_reverse}")

            qpos = query.query_position_or_next

            if query.is_del:
                qpos -= 1

            # print(pileup_col.reference_pos)
            # if q_name == "read_99766/99766/subread/0_SE_97_1160":
            #     print(
            #         pileup_col.reference_pos,
            #         qpos,
            #         query.alignment.query_sequence[
            #             qpos : (qpos + cur_query_max_offset + 1)
            #         ],
            #         query.alignment.is_reverse,
            #     )

            for offset in range(cur_query_max_offset + 1):
                if query.is_del and offset == 0:
                    continue
                self.matrix[self.cur_row + offset, query_pos] = (
                    query.alignment.query_sequence[qpos + offset]
                )

        self.cur_row += max_ins + 1

    def extend_matrix(self, max_offset):
        """if the matrix row have run out, extend the matrix"""
        if (self.cur_row + max_offset) >= self.matrix.shape[0]:
            extended_row = 2 * (self.ref_end - self.cur_ref_pos) + max_offset
            self.matrix = np.concatenate(
                [self.matrix, ResultMatrix.init_matrix(extended_row, self.num_records)],
                axis=0,
            )

    def get_raw_result(self):
        return self.matrix[: self.cur_row]

    def get_query_names(self):
        names = [self.ref_name]
        names.extend(self.query_names)
        return names

    def get_result(self):
        """may the matrix has invalid rows, trim it and return"""
        return self.get_raw_result().transpose()

    def get_result_str(self):
        """valid matrix to string"""
        res = self.get_result()
        names = [self.ref_name]
        names.extend(self.query_names)
        result_strs = []
        for row_idx in range(res.shape[0]):

            q_name = names[row_idx]
            result_strs.append(f">{q_name}")
            result_strs.append("".join(res[row_idx].tolist()))

        return "\n".join(result_strs)

    @staticmethod
    def init_matrix(num_rows, num_cols):
        """init matrix using "." """
        matrix = np.empty(shape=[num_rows, num_cols], dtype=np.str_)
        matrix.fill(".")
        return matrix


def extract_reference(ref_filename):
    """extract reference from fasta file"""
    with open(ref_filename, mode="r", encoding="utf8") as file:
        lines = file.readlines()
        ref_name = lines[0].split(" ")[0][1:]
        return (ref_name.strip(), "".join(lines[1:]))


def bam2fa4jalview(
    aligned_bam_filename,
    ref_filename,
    ref_name,
    interested_ref_start=None,
    interested_ref_end=None,
):
    """generate the fasta info that can be used in jalview from bam file and [ref file]
    Params:
        aligned_bam_filename
        contig
        ref_filename
    """
    samfile = pysam.AlignmentFile(aligned_bam_filename, mode="rb", threads=40)

    fastx_data = Name2Seq(ref_filename)

    ref_seq = fastx_data.fetch(ref_name)

    ref_start = 2**32
    ref_end = 0
    query_names = []
    for query in samfile.fetch(contig=ref_name):
        if (
            interested_ref_start is not None
            and query.reference_end <= interested_ref_start
        ):
            continue
        if (
            interested_ref_end is not None
            and query.reference_start >= interested_ref_end
        ):
            continue
        ref_start = min([query.reference_start, ref_start])
        ref_end = max([query.reference_end, ref_end])

        query_names.append(build_query_name(query))

    interested_ref_start = (
        ref_start if interested_ref_start is None else interested_ref_start
    )
    interested_ref_end = ref_end if interested_ref_end is None else interested_ref_end

    result_matrix = ResultMatrix(
        ref_start=interested_ref_start,
        ref_end=interested_ref_end,
        query_names=query_names,
        ref_name=ref_name,
    )

    for pileup_col in samfile.pileup(contig=ref_name, min_base_quality=1):
        if (
            pileup_col.reference_pos < interested_ref_start
            or pileup_col.reference_pos >= interested_ref_end
        ):
            continue
        result_matrix.update(pileup_col=pileup_col, ref=ref_seq)

    return result_matrix


def plot_msa_align(inp_filename, oup_filename=None):
    """plot the msa align according to the fasta file"""

    from pymsaviz import MsaViz

    mv = MsaViz(
        inp_filename,
        wrap_length=150,
        show_count=True,
        show_grid=True,
        color_scheme="Identity",
    )
    if oup_filename is None:
        oup_filename = f"{inp_filename}.png"
    mv.savefig(oup_filename)


def main(args):
    res = bam2fa4jalview(
        args.bam,
        ref_filename=args.ref_fasta,
        ref_name=args.ref_name,
        interested_ref_start=args.start,
        interested_ref_end=args.end,
    )

    res_str = res.get_result_str()
    if args.o_fasta is not None:
        with open(args.o_fasta, "w", encoding="utf8") as file:
            file.write(res_str)
    else:
        print(res_str)

    if args.o_pic is not None:
        if args.o_fasta is not None:
            plot_msa_align(inp_filename=args.o_fasta, oup_filename=args.o_pic)

        else:
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp:
                tmp.write(res_str)
                tmp.close()
                print(f"temp file name: {tmp.name}")
                plot_msa_align(inp_filename=tmp.name, oup_filename=args.o_pic)
            os.remove(tmp.name)
    return res.first_query_is_rev


def main_cli():
    p = argparse.ArgumentParser(
        "mas_view",
        description="""
        the whole pipeline is:
            1) do alignment (pairwise alignment), minimap2 is an option;
            2) use this script to generate msa pic
    """,
    )
    p.add_argument("--bam", type=str, help="subreads2smc alignment", required=True)
    p.add_argument(
        "--ref-fastx-or-bam",
        type=str,
        help="smc.fa/smc.fq/.bam,",
        required=True,
        dest="ref_fasta",
    )

    p.add_argument("--ref-name", required=True, type=str, dest="ref_name")
    p.add_argument(
        "--ref-start", type=int, default=None, help="contig end", dest="start"
    )
    p.add_argument("--ref-end", type=int, default=None, help="contig start", dest="end")

    p.add_argument(
        "--o-fasta",
        default=None,
        help="output fasta file, if not provided, this content will be output to the stdout",
        dest="o_fasta",
    )
    p.add_argument(
        "--o-pic",
        default=None,
        help="visualization picture file path, if not provided, the plot procedure will be skipped",
        dest="o_pic",
    )
    args = p.parse_args()
    main(args)


if __name__ == "__main__":
    main_cli()



import pysam
import os
import sys

cur_dir = os.path.abspath(__file__).rsplit("/", maxsplit=1)[0]
sys.path.insert(0, cur_dir)

import utils
import polars as pl
import argparse
from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot as plt
import argparse


import polars_init


def main(args):
    # plt.grid(True, linestyle=":", linewidth=0.5, color="gray")

    df = pl.read_csv(args.data, separator="\t")
    df_shift = df.with_columns([pl.col("baseq")])
    df = df.with_columns(
        [
            (pl.col("eq") / (pl.col("eq") + pl.col("diff") + pl.col("ins"))).alias(
                "emp_rq"
            )
        ]
    ).with_columns([utils.q2phreq_expr("emp_rq", "emp_phreq")])
    figure = plt.figure(figsize=(10, 10))
    axs = figure.add_subplot(1, 1, 1)
    plt.sca(axs)
    plt.grid(True, linestyle=":", linewidth=0.5, color="gray")

    sns.scatterplot(df.to_pandas(), x="baseq", y="emp_phreq", ax=axs)
    axs.set_xticks(list(range(0, 60, 2)))
    axs.set_yticks(list(range(0, 60, 2)))
    axs.set_xlabel("PredictedBaseQ", fontdict={"size": 16})
    axs.set_ylabel("EmpericalBaseQ", fontdict={"size": 16})
    perfect_line = pl.DataFrame(
        {
            "x": list(range(0, 60)),
            "y": list(range(0, 60)),
        }
    )

    sns.lineplot(
        perfect_line.to_pandas(), x="x", y="y", ax=axs, color="blue", linestyle="--"
    )

    print(df.head(10))
    figure.savefig(fname="baseq2empq.png")


if __name__ == "__main__":
    polars_init.polars_env_init()
    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("data", metavar="fact_baseq_stat")

    main(parser.parse_args())



import os


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "300"
    os.environ["POLARS_FMT_STR_LEN"] = "100"



import os
import sys

cur_dir = os.path.abspath(__file__).rsplit("/", maxsplit=1)[0]
sys.path.insert(0, cur_dir)

import pysam
import utils
import polars as pl
import argparse
from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot as plt


class BaseQStat:
    def __init__(self, pred_q):
        self.pred_q = pred_q
        self.eq = 0
        self.diff = 0
        self.insertion = 0
        self.deletion = 0
        self.depth = 0

    def add_eq(self, num=1):
        self.eq += num

    def add_diff(self, num=1):
        self.diff += num

    def add_insertion(self, num=1):
        self.insertion += num

    def add_deletion(self, num=1):
        self.deletion += num


def stat_one_record(aligned_pairs, qual, baseq2baseq_stat, query_seq, refseq, ref_end):
    ref_pos_cursor = None
    query_pos_cursor = None

    for qpos, rpos in aligned_pairs:

        if qpos is not None:
            query_pos_cursor = qpos
        if rpos is not None:
            ref_pos_cursor = rpos

        if ref_pos_cursor is None:
            continue

        if query_pos_cursor is None:
            continue

        baseq_stat = baseq2baseq_stat.setdefault(
            qual[query_pos_cursor], BaseQStat(qual[query_pos_cursor])
        )
        assert isinstance(baseq_stat, BaseQStat)
        if rpos is None:
            baseq_stat.add_insertion()
            if ref_pos_cursor == (ref_end - 1):
                break
            continue

        if qpos is None:
            baseq_stat.add_deletion()
        else:
            if refseq[rpos] == query_seq[qpos]:
                baseq_stat.add_eq()
            else:
                baseq_stat.add_diff()

        if ref_pos_cursor == (ref_end - 1):
            break


def stat(aligned_bam_file: str, ref_file: str):
    ref_data = utils.read_fastx_file(ref_file)

    baseq2baseq_stat = {}

    with pysam.AlignmentFile(aligned_bam_file, mode="rb") as bam_h:
        for refname, refseq in ref_data.items():

            for record in tqdm(
                bam_h.fetch(contig=refname), desc=f"processing {refname}"
            ):
                if record.is_secondary or record.is_supplementary or record.is_unmapped:
                    continue

                ref_pos_cursor = None
                query_pos_cursor = None

                ref_start = record.reference_start
                ref_end = record.reference_end
                query_start = record.query_alignment_start
                query_end = record.query_alignment_end

                qual = record.query_qualities
                query_seq = record.query_sequence

                stat_one_record(
                    record.get_aligned_pairs(),
                    qual,
                    baseq2baseq_stat,
                    query_seq,
                    refseq,
                    ref_end,
                )
                # for qpos, rpos in record.get_aligned_pairs():

                #     if qpos is not None:
                #         query_pos_cursor = qpos
                #     if rpos is not None:
                #         ref_pos_cursor = rpos

                #     if ref_pos_cursor is None:
                #         continue

                #     if query_pos_cursor is None:
                #         continue

                #     baseq_stat = baseq2baseq_stat.setdefault(
                #         qual[query_pos_cursor], BaseQStat(qual[query_pos_cursor])
                #     )
                #     assert isinstance(baseq_stat, BaseQStat)
                #     if rpos is None:
                #         baseq_stat.add_insertion()
                #         if ref_pos_cursor == (ref_end - 1):
                #             break
                #         continue

                #     if qpos is None:
                #         baseq_stat.add_deletion()
                #     else:
                #         if refseq[rpos] == query_seq[qpos]:
                #             baseq_stat.add_eq()
                #         else:
                #             baseq_stat.add_diff()

                #     if ref_pos_cursor == (ref_end - 1):
                #         break
    baseqs = []
    eqs = []
    diffs = []
    insertions = []
    deletions = []

    for bq, stat in baseq2baseq_stat.items():
        assert isinstance(stat, BaseQStat)
        baseqs.append(bq)
        eqs.append(stat.eq)
        diffs.append(stat.diff)
        insertions.append(stat.insertion)
        deletions.append(stat.deletion)

    return pl.DataFrame(
        {"baseq": baseqs, "eq": eqs, "diff": diffs, "ins": insertions, "del": deletions}
    )


def main(args):
    df = stat(args.aln_bam, args.ref_file)
    df = df.with_columns(
        [
            (pl.col("eq") / (pl.col("eq") + pl.col("diff") + pl.col("ins"))).alias(
                "emp_rq"
            )
        ]
    ).with_columns([utils.q2phreq_expr("emp_rq", "emp_phreq")])
    figure = plt.figure(figsize=(20, 10))
    axs = figure.add_subplot(1, 1, 1)
    sns.scatterplot(df.to_pandas(), x="baseq", y="emp_phreq", ax=axs)

    print(df.head(10))
    figure.savefig(fname="baseq2empq")


if __name__ == "__main__":

    params = {
        "aln_bam": "/data/ccs_data/ccs_eval2024q4/output-all/smc2ref.bam",
        "ref_file": "/data/ccs_data/MG1655.fa",
    }

    main(argparse.Namespace(**params))



import subprocess
import pathlib
import os
import logging
import polars as pl
import shutil
from multiprocessing import cpu_count

logging.basicConfig(
    level=logging.INFO,
    datefmt="%Y/%m/%d %H:%M:%S",
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def extract_filename(filepath: str) -> str:
    p = pathlib.Path(filepath)
    return p.stem


def do_alignment(
    bam_file: str, ref_fasta: str, outdir: str, force: bool = False, threads=None
) -> str:

    res_bam_prefix = "{}/{}.aligned".format(outdir, extract_filename(bam_file))
    result_bam = f"{res_bam_prefix}.bam"

    if force and os.path.exists(result_bam):
        os.remove(result_bam)

    if os.path.exists(result_bam):
        logging.info(f"{result_bam} exists, use it now")
        return result_bam
    threads = cpu_count() if threads is None else threads
    cmd = f"""gsmm2 --threads {threads} align \
            -q {bam_file} \
            -t {ref_fasta} \
            -p {res_bam_prefix} \
            --kmer 11 \
            --wins 1 \
            --noSeco"""

    logging.info("cmd: %s", cmd)
    subprocess.check_call(cmd, shell=True)

    return result_bam


def generate_fact_table(aligned_bam: str, ref_fasta: str, outdir: str):
    """_summary_

    Args:
        aligned_bam (str): aligned bam
        outdir (str): the outdir must be empty or not in use
    """
    if os.path.exists(outdir):
        raise ValueError(f"{outdir} already exists")
    cmd = f"""gsetl -f --outdir {outdir} \
        aligned-bam \
        --bam {aligned_bam} \
        --ref-file {ref_fasta} \
        --factRefLocusInfo 0 \
        --factErrorQueryLocusInfo 0 \
        --factBaseQStat 0 \
        --factPolyInfo 0 \
        --factRecordStat 0 \
        --useSupp
        """
    logging.info("cmd: %s", cmd)
    subprocess.check_call(cmd, shell=True)

    fact_bam_basic = f"{outdir}/fact_aligned_bam_bam_basic.csv"
    # fact_bam_record_stat = f"{outdir}/fact_aligned_bam_record_stat.csv"
    return fact_bam_basic, None


# def stats(fact_bam_basic, fact_bam_record_stat, file_h):
#     fact_bam_basic = pl.read_csv(fact_bam_basic, separator="\t")
#     fact_bam_record_stat = pl.read_csv(fact_bam_record_stat, separator="\t")
#     # print(fact_bam_basic.head(2))
#     # print(fact_bam_record_stat.head(2))

#     joined = fact_bam_basic.join(fact_bam_record_stat, on="qname")
#     # print(joined.head(2))
#     query_coverage = joined.select(
#         [
#             pl.col("qlen"),
#             (
#                 pl.col("matchBp")
#                 + pl.col("mismatchBp")
#                 + pl.col("nonHpInsertionBp")
#                 + pl.col("hpInsertionBp")
#             ).alias("alignedBp"),
#         ]
#     ).select(
#         [(pl.col("alignedBp").sum() / pl.col("qlen").sum()).alias("queryCoverage")]
#     )
#     qc = query_coverage.to_numpy()[0][0]

#     file_h.write(f"queryCoverage\t{qc}\n")
#     # print(query_coverage)


def stats(fact_bam_basic, file_h):
    fact_bam_basic = pl.read_csv(fact_bam_basic, separator="\t")
    df = (
        fact_bam_basic.select(
            [
                pl.col("qname"),
                pl.col("qlen"),
                pl.when(pl.col("fwd"))
                .then(pl.col("qstart"))
                .otherwise(pl.col("qlen") - pl.col("qend"))
                .alias("qstart"),
                pl.when(pl.col("fwd"))
                .then(pl.col("qend"))
                .otherwise(pl.col("qlen") - pl.col("qstart"))
                .alias("qend"),
            ]
        )
        .select(
            [pl.col("qname"), pl.col("qlen"), pl.struct("qstart", "qend").alias("se")]
        )
        .group_by("qname")
        .agg([pl.col("qlen").max(), pl.col("se"), pl.len().alias("cnt")])
    )

    # print(df.filter(pl.col("cnt") > 1).count())
    # os._exit(0)

    df = df.with_columns(
        [
            pl.col("se")
            .map_elements(
                lambda x: merge_intervals(x),
                return_dtype=pl.UInt64,
                returns_scalar=True,
            )
            .alias("aligned_qlen")
        ]
    )

    # print(df.filter(pl.col("aligned_qlen") < 0).head(2))
    # print(df.head(2))

    df = df.select(
        [
            (pl.col("aligned_qlen").sum() / pl.col("qlen").cast(pl.UInt64).sum()).alias(
                "queryCoverage"
            )
        ]
    )
    print(df)
    qc = df.to_numpy()[0][0]
    file_h.write(f"queryCoverage\t{qc}\n")


def merge_intervals(region_list):
    # print(region_list)
    region_list = [[se["qstart"], se["qend"]] for se in region_list]
    if len(region_list) == 1:
        return region_list[0][1] - region_list[0][0]

    # Step 1: Sort intervals by start position
    sorted_regions = sorted(region_list, key=lambda x: x[0])

    # Step 2: Merge overlapping intervals
    merged_regions = []
    current_start, current_end = sorted_regions[0][0], sorted_regions[0][1]

    for start, end in sorted_regions[1:]:
        if start <= current_end:  # Overlapping or adjacent regions
            current_end = max(current_end, end)  # Extend the current region
        else:
            # No overlap, so save the current region and start a new one
            merged_regions.append((current_start, current_end))
            current_start, current_end = start, end

    # Add the last region
    merged_regions.append((current_start, current_end))

    # Step 3: Calculate the total length of merged intervals
    total_length = sum([(se[1] - se[0]) for se in merged_regions])
    # print(merged_regions, total_length)
    # if total_length < 0:
    #     print(merge_intervals, total_length)

    return total_length


def main(bam_file: str, ref_fa: str, threads=None, force=False, outdir=None) -> str:
    """
        step1: do alignment
        step2: generate detailed metric info
        step3: compute the aggr metric. the result aggr_metric.csv is a '\t' seperated csv file. the header is name\tvalue
            here is a demo.
            ---------aggr_metric.csv
            name    value
            queryCoverage   0.937
            ----------

    requirements:
        mm2: cargo install mm2
        gsetl: cargo install gsetl

    Args:
        bam_file (str): bam file. only support adapter.bam
        ref_fa (str): ref genome fa file nam
        force (boolean): if force==False, the outdir must not exists in advance. if force==True, the outdir will be removed if exists
            the proceduer will create a empty outdir for the metric related files
        outdir:
            if outdir provided, read ${outdir}/metric/aggr_metric.csv for metric result
            if not, read ${bam_file_dir}/${bam_file_name}-metric/metric/aggr_metric.csv for metric result

    Return:
        aggr_metric_filename (str): the aggr metric file
    """
    bam_filedir = os.path.dirname(bam_file)
    bam_filename = extract_filename(bam_file)
    if outdir is None:
        outdir = os.path.join(bam_filedir, f"{bam_filename}-metric")
    if force and os.path.exists(outdir):
        shutil.rmtree(outdir)
    if os.path.exists(outdir):
        raise ValueError(
            f"{outdir} already exists. remove it by manually or force=True"
        )

    os.makedirs(outdir)

    metric_outdir = os.path.join(outdir, "metric")

    aligned_bam_file = do_alignment(
        bam_file, ref_fa, outdir=outdir, force=force, threads=threads
    )

    fact_bam_basic, _ = generate_fact_table(
        aligned_bam_file, ref_fasta=ref_fa, outdir=metric_outdir
    )

    aggr_metric_filename = os.path.join(metric_outdir, "aggr_metric.csv")

    with open(aggr_metric_filename, encoding="utf8", mode="w") as file_h:
        file_h.write(f"name\tvalue\n")
        stats(fact_bam_basic, file_h=file_h)
    return aggr_metric_filename


def test_stat():
    fact_bam_basic = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/fact_aligned_bam_bam_basic.csv"
    aggr_metric_filename = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/aggr_metric.csv"

    with open(aggr_metric_filename, encoding="utf8", mode="w") as file_h:
        file_h.write(f"name\tvalue\n")
        stats(fact_bam_basic, file_h=file_h)


if __name__ == "__main__":
    bam_files = [
        "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter.bam",
        "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0002_adapter.bam",
        "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0003_adapter.bam",
    ]
    ref = "/data/ccs_data/MG1655.fa"

    for bam in bam_files:
        main(bam_file=bam, ref_fa=ref, force=True)
    # test_stat()

    # print(merge_intervals([{"qstart": 11, "qend": 771}]))



import pysam
from typing import Mapping, Tuple
from tqdm import tqdm
import polars as pl


def read_bam_file(bam_file: str) -> Mapping[str, Tuple[str, int]]:
    res = {}
    with pysam.AlignmentFile(bam_file, mode="rb", threads=40, check_sq=False) as bam_h:
        for record in tqdm(
            bam_h.fetch(until_eof=True), desc=f"read_bam_file:>> reading {bam_file}"
        ):
            res[record.query_name] = (record.query_sequence, f"00_{record.query_name}")

    return res


def read_fastx_file(fname: str) -> Mapping[str, Tuple[str, int]]:
    fh = pysam.FastxFile(fname)
    res = {}
    for entry in fh:
        res[entry.name] = (entry.sequence, f"00_{entry.name}")
    return res


def q2phreq_expr(inp_name, oup_name=None):
    oup_name = oup_name if oup_name is not None else inp_name
    return (
        -10.0
        * (
            1
            - pl.when(pl.col(inp_name) > (1 - 1e-6))
            .then(1 - 1e-6)
            .otherwise(pl.col(inp_name))
        ).log10()
    ).alias(oup_name)
